# -*- coding: utf-8 -*-
#
#   DIM-SDK : Decentralized Instant Messaging Software Development Kit
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Dialog Robot
    ~~~~~~~~~~~~

    Dialog for chatting with station
"""

import random
from typing import Union

from dimp import ID
from dimp import Content, TextContent, AudioContent

from .chatbot import ChatBot


class Dialog:

    def __init__(self):
        super().__init__()
        # chat bot candidates
        self.__bots: list = []

    @property
    def bots(self) -> list:
        return self.__bots

    @bots.setter
    def bots(self, array: Union[list, ChatBot]):
        if isinstance(array, list):
            count = len(array)
            if count > 1:
                # set bots with random order
                self.__bots = random.sample(array, len(array))
            else:
                self.__bots = array
        elif isinstance(array, ChatBot):
            self.__bots = [array]
        else:
            raise ValueError('bots error: %s' % array)

    def ask(self, question: str, sender: ID) -> str:
        # try each chat robots
        index = 0
        for robot in self.__bots:
            answer = robot.ask(question=question, user=sender.number)
            if answer is None:
                index += 1
                continue
            # got the answer
            if index > 0:
                # move this bot to front
                self.__bots.remove(robot)
                self.__bots.insert(0, robot)
            return answer

    def query(self, content: Content, sender: ID) -> TextContent:
        if isinstance(content, TextContent):
            # text dialog
            question = content.text
            answer = self.ask(question=question, sender=sender)
            if answer is not None:
                response = TextContent.new(text=answer)
                group = content.group
                if group is not None:
                    response.group = group
                return response
        elif isinstance(content, AudioContent):
            # TODO: Automatic Speech Recognition
            pass



# -*- coding: utf-8 -*-
#
#   DIM-SDK : Decentralized Instant Messaging Software Development Kit
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Certificate Authority
    ~~~~~~~~~~~~~~~~~~~~~

    Certificates for Stations and Service Providers
"""

import json

from mkm.crypto.utils import base64_decode

from dimp import PublicKey


class CASubject(dict):

    # def __new__(cls, subject: dict):
    #     return super().__new__(cls, subject)

    def __init__(self, subject: dict):
        super().__init__(subject)
        if 'C' in subject:
            self.country = subject['C']   # C:  CN, US, ...
        else:
            self['C'] = 'CN'
            self.country = 'CN'
        if 'ST' in subject:
            self.state = subject['ST']    # ST: province
        else:
            self.state = None
        if 'L' in subject:
            self.locality = subject['L']  # L:  city
        else:
            self.locality = None

        self.organization = subject['O']  # O:  Co., Ltd.
        self.unit = subject['OU']         # OU: Department
        self.name = subject['CN']         # CN: domain/ip


class CAValidity(dict):

    def __init__(self, validity: dict):
        super().__init__(validity)
        self.begin = int(validity['NotBefore'])  # notBefore
        self.end = int(validity['NotAfter'])     # notAfter


class CAData(dict):

    def __init__(self, data: dict):
        super().__init__(data)
        self.issuer = CASubject(data['Issuer'])       # issuer DN
        self.validity = CAValidity(data['Validity'])  # validity
        self.subject = CASubject(data['Subject'])     # the CA owner
        if 'PublicKey' in data:
            self.key = PublicKey(data['PublicKey'])   # owner's PK
        elif 'Key' in data:
            self.key = PublicKey(data['Key'])


class CertificateAuthority(dict):

    def __init__(self, ca: dict):
        super().__init__(ca)
        # version
        self.version = int(ca['version'])
        # serial number
        if 'serialNumber' in ca:
            self.sn = ca['serialNumber']
        else:
            self.sn = ca['sn']
        # info (JsON string)
        self.info = CAData(json.loads(ca['info']))
        # signature of info with Issuer's Public Key
        self.signature = base64_decode(ca['signature'])
        # extensions (dict)
        if 'extensions' in ca:
            self.extensions = ca['extensions']



# -*- coding: utf-8 -*-
#
#   DIM-SDK : Decentralized Instant Messaging Software Development Kit
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Messenger
    ~~~~~~~~~

    Transform and send message
"""

import weakref
from abc import abstractmethod
from typing import Optional, Union

from dimp import SymmetricKey, ID, Meta, User
from dimp import InstantMessage, SecureMessage, ReliableMessage
from dimp import Content, FileContent
from dimp import Transceiver

from .delegate import Callback, CompletionHandler
from .delegate import MessengerDelegate, ConnectionDelegate
from .facebook import Facebook

from .cpu import ContentProcessor


class Messenger(Transceiver, ConnectionDelegate):

    def __init__(self):
        super().__init__()
        self.__context: dict = {}
        self.__delegate: weakref.ReferenceType = None
        self.__cpu = ContentProcessor(messenger=self)

    #
    #   Delegate for sending data
    #
    @property
    def delegate(self) -> Optional[MessengerDelegate]:
        if self.__delegate is not None:
            return self.__delegate()

    @delegate.setter
    def delegate(self, value: Optional[MessengerDelegate]):
        assert value is not None, 'message delegate should not be empty'
        self.__delegate = weakref.ref(value)

    #
    #   Environment variables as context
    #
    @property
    def context(self) -> dict:
        return self.__context

    def get_context(self, key: str):
        return self.__context.get(key)

    def set_context(self, key: str, value):
        if value is None:
            self.__context.pop(key, None)
        else:
            self.__context[key] = value

    #
    #   Data source for getting entity info
    #
    @property
    def facebook(self) -> Facebook:
        barrack = self.__context.get('facebook')
        if barrack is None:
            barrack = self.barrack
            assert isinstance(barrack, Facebook), 'messenger delegate error: %s' % barrack
        return barrack

    def __select(self, receiver: ID) -> Optional[User]:
        facebook = self.facebook
        users = facebook.local_users
        if users is None or len(users) == 0:
            raise LookupError('local users should not be empty')
        elif receiver.is_broadcast:
            # broadcast message can decrypt by anyone, so just return current user
            return users[0]
        if receiver.is_group:
            # group message (recipient not designated)
            for item in users:
                if facebook.exists_member(member=item.identifier, group=receiver):
                    # set this item to be current user?
                    return item
        else:
            # 1. personal message
            # 2. split group message
            assert receiver.is_user, 'receiver ID error: %s' % receiver
            for item in users:
                if item.identifier == receiver:
                    # set this item to be current user?
                    return item

    def __trim(self, msg: SecureMessage) -> Optional[SecureMessage]:
        receiver = self.facebook.identifier(msg.envelope.receiver)
        user = self.__select(receiver=receiver)
        if user is None:
            # current users not match
            msg = None
        elif receiver.is_group:
            # trim group message
            msg = msg.trim(member=user.identifier)
        return msg

    #
    #  Transform
    #
    def verify_message(self, msg: ReliableMessage) -> Optional[SecureMessage]:
        facebook = self.facebook
        # NOTICE: check meta before calling me
        sender = facebook.identifier(msg.envelope.sender)
        meta = msg.meta
        if meta is None:
            meta = facebook.meta(identifier=sender)
            if meta is None:
                # NOTICE: the application will query meta automatically
                # save this message in a queue waiting sender's meta response
                self.suspend_message(msg=msg)
                # raise LookupError('failed to get meta for sender: %s' % sender)
                return None
        else:
            # [Meta Protocol]
            # save meta for sender
            meta = Meta(meta)
            if not facebook.save_meta(meta=meta, identifier=sender):
                raise ValueError('save meta error: %s, %s' % (sender, meta))
        return super().verify_message(msg=msg)

    def encrypt_message(self, msg: InstantMessage) -> SecureMessage:
        s_msg = super().encrypt_message(msg=msg)
        group = msg.content.group
        if group is not None:
            # NOTICE: this help the receiver knows the group ID
            #         when the group message separated to multi-messages,
            #         if don't want the others know you are the group members,
            #         remove it.
            s_msg.envelope.group = group
        # NOTICE: copy content type to envelope
        #         this help the intermediate nodes to recognize message type
        s_msg.envelope.type = msg.content.type
        return s_msg

    def decrypt_message(self, msg: SecureMessage) -> Optional[InstantMessage]:
        # trim message
        s_msg = self.__trim(msg=msg)
        if s_msg is None:
            # not for you?
            raise LookupError('receiver error: %s' % msg)
        # decrypt message
        return super().decrypt_message(msg=s_msg)

    #
    #   InstantMessageDelegate
    #
    def encrypt_content(self, content: Content, key: dict, msg: InstantMessage) -> bytes:
        password = SymmetricKey(key=key)
        assert password == key, 'irregular symmetric key: %s' % key
        # check attachment for File/Image/Audio/Video message content before
        if isinstance(content, FileContent):
            data = password.encrypt(data=content.data)
            # upload (encrypted) file data onto CDN and save the URL in message content
            url = self.delegate.upload_data(data=data, msg=msg)
            if url is not None:
                content.url = url
                content.data = None
        return super().encrypt_content(content=content, key=password, msg=msg)

    def encrypt_key(self, key: dict, receiver: str, msg: InstantMessage) -> Optional[bytes]:
        facebook = self.facebook
        to = facebook.identifier(receiver)
        pk = facebook.public_key_for_encryption(identifier=to)
        if pk is None:
            meta = facebook.meta(identifier=to)
            if meta is None:
                # save this message in a queue waiting receiver's meta response
                self.suspend_message(msg=msg)
                # raise LookupError('failed to get encrypt key for receiver: %s' % receiver)
                return None
        return super().encrypt_key(key=key, receiver=receiver, msg=msg)

    #
    #   SecureMessageDelegate
    #
    def decrypt_content(self, data: bytes, key: dict, msg: SecureMessage) -> Optional[Content]:
        password = SymmetricKey(key=key)
        content = super().decrypt_content(data=data, key=password, msg=msg)
        if content is None:
            return None
        # check attachment for File/Image/Audio/Video message content after
        if isinstance(content, FileContent):
            i_msg = InstantMessage.new(content=content, envelope=msg.envelope)
            # download from CDN
            file_data = self.delegate.download_data(content.url, i_msg)
            if file_data is None:
                # save symmetric key for decrypted file data after download from CDN
                content.password = password
            else:
                # decrypt file data
                content.data = password.decrypt(data=file_data)
                assert content.data is not None, 'failed to decrypt file data with key: %s' % key
                content.url = None
        return content

    #
    #   Send message
    #
    def send_content(self, content: Content, receiver: ID, callback: Callback=None, split: bool=True) -> bool:
        """
        Send content to receiver

        :param content: message content
        :param receiver: receiver ID
        :param callback: callback function
        :param split:    if it's a group message, split it before sending out
        :return: True on success
        """
        user = self.facebook.current_user
        assert user is not None, 'failed to get current user'
        i_msg = InstantMessage.new(content=content, sender=user.identifier, receiver=receiver)
        return self.send_message(msg=i_msg, callback=callback, split=split)

    def send_message(self, msg: Union[InstantMessage, ReliableMessage],
                     callback: Callback=None, split: bool=True) -> bool:
        """
        Send instant message (encrypt and sign) onto DIM network

        :param msg:      instant message
        :param callback: callback function
        :param split:    if it's a group message, split it before sending out
        :return:         False on data/delegate error
        """
        if isinstance(msg, ReliableMessage):
            return self.__send_message(msg=msg, callback=callback)
        elif not isinstance(msg, InstantMessage):
            raise TypeError('message error: %s' % msg)
        facebook = self.facebook
        # Send message (secured + certified) to target station
        s_msg = self.encrypt_message(msg=msg)
        r_msg = self.sign_message(msg=s_msg)
        receiver = facebook.identifier(msg.envelope.receiver)
        ok = True
        if split and receiver.is_group:
            # split for each members
            members = facebook.members(identifier=receiver)
            if members is None or len(members) == 0:
                # FIXME: query group members from sender
                messages = None
            else:
                messages = r_msg.split(members=members)
            if messages is None:
                # failed to split msg, send it to group
                ok = self.__send_message(msg=r_msg, callback=callback)
            else:
                # sending group message one by one
                for r_msg in messages:
                    if not self.__send_message(msg=r_msg, callback=callback):
                        ok = False
        else:
            ok = self.__send_message(msg=r_msg, callback=callback)
        # TODO: if OK, set iMsg.state = sending; else set iMsg.state = waiting

        if not self.save_message(msg=msg):
            return False
        return ok

    def __send_message(self, msg: ReliableMessage, callback: Callback) -> bool:
        handler = MessageCallback(msg=msg, cb=callback)
        data = self.serialize_message(msg=msg)
        return self.delegate.send_package(data=data, handler=handler)

    #
    #   Saving Message
    #

    @abstractmethod
    def save_message(self, msg: InstantMessage) -> bool:
        """
        Save instant message

        :param msg: instant message
        :return: True on success
        """
        raise NotImplemented

    @abstractmethod
    def suspend_message(self, msg: Union[ReliableMessage, InstantMessage]):
        """
        Suspend reliable message for the sender's meta
        Suspend instant message for the receiver's meta or group info

        :param msg: message received from network / instant message to be sent
        """
        # NOTICE: this function is for Client
        #         if the client cannot get verify/encrypt message for contact,
        #         it means you should suspend it and query meta from DIM station first
        raise NotImplemented

    #
    #   ConnectionDelegate
    #
    def received_package(self, data: bytes) -> Optional[bytes]:
        """
        Processing received message package

        :param data: message data
        :return: response message data
        """
        # 1. deserialize message
        r_msg = self.deserialize_message(data=data)
        if r_msg is None:
            # no message received
            return None
        # 2. process message
        r_msg = self.process_reliable(msg=r_msg)
        if r_msg is None:
            # nothing to response
            return None
        # 3. serialize message
        return self.serialize_message(msg=r_msg)

    # TODO: override to check broadcast message before calling it
    # TODO: override to deliver to the receiver when catch exception "receiver error ..."
    def process_reliable(self, msg: ReliableMessage) -> Optional[ReliableMessage]:
        # 1. verify message
        s_msg = self.verify_message(msg=msg)
        if s_msg is None:
            # waiting for sender's meta if not exists
            return None
        # 2. process message
        s_msg = self.process_secure(msg=s_msg)
        if s_msg is None:
            # nothing to respond
            return None
        # 3. sign message
        return self.sign_message(msg=s_msg)

    def process_secure(self, msg: SecureMessage) -> Optional[SecureMessage]:
        # 1. decrypt message
        i_msg = self.decrypt_message(msg=msg)
        if i_msg is None:
            # cannot decrypt this message, not for you?
            # delivering message to other receiver?
            return None
        # 2. process message
        i_msg = self.process_instant(msg=i_msg)
        if i_msg is None:
            # nothing to respond
            return None
        # 3. encrypt message
        return self.encrypt_message(msg=i_msg)

    # TODO: override to check group
    # TODO: override to filter the response
    def process_instant(self, msg: InstantMessage) -> Optional[InstantMessage]:
        facebook = self.facebook
        sender = facebook.identifier(string=msg.envelope.sender)
        # process content from sender
        res = self.__cpu.process(content=msg.content, sender=sender, msg=msg)
        if not self.save_message(msg=msg):
            # error
            return None
        if res is None:
            # nothing to respond
            return None
        # check receiver
        receiver = facebook.identifier(msg.envelope.receiver)
        user = self.__select(receiver=receiver)
        assert user is not None, 'receiver error: %s' % receiver
        # pack message
        return InstantMessage.new(content=res, sender=user.identifier, receiver=sender)


class MessageCallback(CompletionHandler):

    def __init__(self, msg: ReliableMessage, cb: Callback):
        super().__init__()
        self.msg = msg
        self.callback = cb

    def success(self):
        if self.callback is not None:
            self.callback.finished(result=self.msg, error=None)

    def failed(self, error):
        if self.callback is not None:
            self.callback.finished(result=self.msg, error=error)



# -*- coding: utf-8 -*-
#
#   DIM-SDK : Decentralized Instant Messaging Software Development Kit
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Meta Command Processor
    ~~~~~~~~~~~~~~~~~~~~~~

"""

from typing import Optional

from dimp import ID, Meta
from dimp import InstantMessage
from dimp import Content
from dimp import TextContent, Command, MetaCommand

from ..protocol import ReceiptCommand

from .command import CommandProcessor


class MetaCommandProcessor(CommandProcessor):

    def __get(self, identifier: ID) -> Content:
        # query meta for ID
        meta = self.facebook.meta(identifier=identifier)
        if meta is None:
            # meta not found
            return TextContent.new(text='Sorry, meta for %s not found.' % identifier)
        # response
        return MetaCommand.response(identifier=identifier, meta=meta)

    def __put(self, identifier: ID, meta: Meta) -> Content:
        facebook = self.facebook
        # received a meta for ID
        if not facebook.verify_meta(meta=meta, identifier=identifier):
            # meta not match
            return TextContent.new(text='Meta not match ID: %s' % identifier)
        if not facebook.save_meta(meta=meta, identifier=identifier):
            # save meta failed
            return TextContent.new(text='Meta not accept: %s!' % identifier)
        # response
        return ReceiptCommand.new(message='Meta received: %s' % identifier)

    #
    #   main
    #
    def process(self, content: Content, sender: ID, msg: InstantMessage) -> Optional[Content]:
        assert isinstance(content, MetaCommand), 'command error: %s' % content
        identifier = self.facebook.identifier(content.identifier)
        meta = content.meta
        if meta is None:
            return self.__get(identifier=identifier)
        else:
            return self.__put(identifier=identifier, meta=meta)


# register
CommandProcessor.register(command=Command.META, processor_class=MetaCommandProcessor)



# -*- coding: utf-8 -*-
#
#   DIM-SDK : Decentralized Instant Messaging Software Development Kit
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================


def read_int(data, pos):
    return int.from_bytes(bytes=data[pos:pos + 4], byteorder='big')


def append_int(data, i: int):
    return data + i.to_bytes(length=4, byteorder='big')


MIN_HEAD_LEN = 4 + 4 + 4 + 4 + 4


class NetMsgHead(bytes):

    SEND_MSG = 3
    NOOP = 6

    """
        Message Packer for Tencent/mars
        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        data structure:
            head_length - 4 bytes
            version     - 4 bytes
            cmd         - 4 bytes
            seq         - 4 bytes
            body_len    - 4 bytes
            options     - variable length (head_length - 20)
    """

    def __new__(cls, data: bytes=None,
                version: int = 200, cmd: int=0, seq: int=0, options: bytes=None, body: bytes=None):
        """

        :param data:    msg pack head data (if data is None, use other parameters to create msg head)
        :param version: client version
        :param cmd:     cmd id
        :param seq:     serial number
        :param options: extra parameters
        :param body:    msg body
        :return:
        """
        if data:
            # return NetMsgHead object directly
            if isinstance(data, NetMsgHead):
                return data
            # check data length
            data_len = len(data)
            if data_len < MIN_HEAD_LEN:
                raise ValueError('net message pack data length error: %d' % data_len)
            # get fields
            head_len = read_int(data, 0)
            version = read_int(data, 4)
            cmd = read_int(data, 8)
            seq = read_int(data, 12)
            body_len = read_int(data, 16)
            # check head length
            if data_len < head_len or head_len < MIN_HEAD_LEN:
                raise ValueError('net message pack head length error: %d' % head_len)
            elif data_len > head_len:
                # cut head
                data = data[:head_len]
            # get options
            if head_len == MIN_HEAD_LEN:
                options = None
            else:
                options = data[MIN_HEAD_LEN:]
            # check body length
            if body_len < 0:
                raise ValueError('net message pack body length error: %d' % body_len)
        else:
            if options:
                head_len = MIN_HEAD_LEN + len(options)
            else:
                head_len = MIN_HEAD_LEN
            if body:
                body_len = len(body)
            else:
                body_len = 0
            data = b''
            data = append_int(data, head_len)
            data = append_int(data, version)
            data = append_int(data, cmd)
            data = append_int(data, seq)
            data = append_int(data, body_len)
            if options:
                data = data + options
        # create net message head
        self = super().__new__(cls, data)
        self.head_length = head_len
        self.version = version
        self.cmd = cmd
        self.seq = seq
        self.body_length = body_len
        self.options = options
        return self


class NetMsg(bytes):

    def __new__(cls, data: bytes=None,
                head: bytes=None,
                version: int=200, cmd: int=0, seq: int=0, options: bytes=None, body: bytes=None):
        """

        :param data:    msg pack data (if data is None, use other parameters to create msg pack)
        :param head:    msg pack head data (if head is None, use other parameters to create msg head)
        :param version:
        :param cmd:
        :param seq:
        :param options:
        :param body:
        :return:
        """
        if data:
            # return NetMsg object directly
            if isinstance(data, NetMsg):
                return data
            # check head
            if head is None:
                head = NetMsgHead(data)
            # check pack length
            pack_len = head.head_length + head.body_length
            data_len = len(data)
            if data_len < pack_len:
                raise ValueError('data length error: %d < %d' % (data_len, pack_len))
            elif data_len > pack_len:
                # cut package
                data = data[:pack_len]
            # get body
            if head.body_length == 0:
                body = None
            else:
                body = data[head.head_length:]
        elif head:
            head = NetMsgHead(head)
            if body:
                if head.body_length == len(body):
                    data = head + data
                else:
                    raise ValueError('body length error:', head.body_length)
            else:
                data = head
        else:
            # if head is None, use other parameters to create head
            head = NetMsgHead(version=version, cmd=cmd, seq=seq, options=options, body=body)
            if body:
                data = head + body
            else:
                data = head
        # create net message package
        self = super().__new__(cls, data)
        self.head = head
        self.body = body
        return self



# -*- coding: utf-8 -*-
#
#   DIM-SDK : Decentralized Instant Messaging Software Development Kit
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from dimp import *

from .notification import Notification, Observer, NotificationCenter
from .notification import ApplePushNotificationService

from .nlp import Dialog, ChatBot, Tuling, XiaoI

from .network import NetMsgHead, NetMsg
from .network import Session, SessionServer
from .network import CASubject, CAValidity, CAData, CertificateAuthority
from .network import ServiceProvider, Station, Robot

from .group import Polylogue

from .protocol import ReceiptCommand
from .protocol import BlockCommand, MuteCommand
from .protocol import StorageCommand

from .cpu import ContentProcessor, ForwardContentProcessor
from .cpu import CommandProcessor, HistoryCommandProcessor, GroupCommandProcessor
from .cpu import InviteCommandProcessor, ExpelCommandProcessor, QuitCommandProcessor
from .cpu import ResetCommandProcessor, QueryCommandProcessor
from .cpu import MetaCommandProcessor, ProfileCommandProcessor

from .delegate import Callback, CompletionHandler, MessengerDelegate, ConnectionDelegate

from .ans import AddressNameService
from .facebook import Facebook
from .keystore import KeyStore
from .messenger import Messenger

name = 'DIM-SDK'

__author__ = 'Albert Moky'

__all__ = [

    #
    #   MingKeMing
    #

    # crypto
    'SignKey', 'VerifyKey', 'EncryptKey', 'DecryptKey',
    'SymmetricKey', 'PrivateKey', 'PublicKey',

    # entity
    'NetworkID', 'Address', 'ID', 'Meta', 'Profile',
    'Entity', 'User', 'Group',

    # delegate
    'EntityDataSource', 'UserDataSource', 'GroupDataSource',

    'ANYONE', 'EVERYONE',

    #
    #   DaoKeDao
    #

    # message
    'Envelope', 'Content', 'Message',

    # content types
    'ContentType', 'ForwardContent',

    # transform
    'InstantMessage', 'SecureMessage', 'ReliableMessage',

    # delegate
    'InstantMessageDelegate', 'SecureMessageDelegate', 'ReliableMessageDelegate',

    #
    #   DIMP
    #

    # protocol
    'TextContent', 'FileContent', 'ImageContent', 'AudioContent', 'VideoContent',
    'Command', 'HistoryCommand', 'GroupCommand',
    'InviteCommand', 'ExpelCommand', 'JoinCommand', 'QuitCommand',
    'QueryCommand', 'ResetCommand',
    'HandshakeCommand', 'MetaCommand', 'ProfileCommand',

    # core
    'Barrack', 'KeyCache', 'Transceiver',

    # delegate
    'EntityDelegate', 'CipherKeyDelegate',

    #
    #   DIM SDK
    #

    # notification
    'Notification', 'Observer', 'NotificationCenter',
    'ApplePushNotificationService',

    # NLP
    'Dialog', 'ChatBot', 'Tuling', 'XiaoI',

    # network
    'NetMsgHead', 'NetMsg',
    'Session', 'SessionServer',
    'CASubject', 'CAValidity', 'CAData', 'CertificateAuthority',
    'ServiceProvider', 'Station', 'Robot',

    # group
    'Polylogue',

    # protocol
    'ReceiptCommand',
    'BlockCommand', 'MuteCommand',
    'StorageCommand',

    # cpu
    'ContentProcessor', 'ForwardContentProcessor',
    'CommandProcessor', 'HistoryCommandProcessor', 'GroupCommandProcessor',
    'InviteCommandProcessor', 'ExpelCommandProcessor', 'QuitCommandProcessor',
    'ResetCommandProcessor', 'QueryCommandProcessor',
    'MetaCommandProcessor', 'ProfileCommandProcessor',

    # delegate
    'Callback', 'CompletionHandler', 'MessengerDelegate', 'ConnectionDelegate',

    'AddressNameService', 'Facebook', 'KeyStore', 'Messenger',
]



# -*- coding: utf-8 -*-
#
#   DIM-SDK : Decentralized Instant Messaging Software Development Kit
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Query Group Command Processor
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    1. query for group members-list
    2. any existed member or assistant can query group members-list
"""

from typing import Optional

from dimp import ID
from dimp import InstantMessage
from dimp import Content, TextContent
from dimp import GroupCommand, QueryCommand

from .history import GroupCommandProcessor


class QueryCommandProcessor(GroupCommandProcessor):

    #
    #   main
    #
    def process(self, content: Content, sender: ID, msg: InstantMessage) -> Optional[Content]:
        assert isinstance(content, QueryCommand), 'group command error: %s' % content
        facebook = self.facebook
        group: ID = facebook.identifier(content.group)
        # 1. check permission
        if not facebook.exists_member(member=sender, group=group):
            if not facebook.exists_assistant(member=sender, group=group):
                if not facebook.is_owner(member=sender, group=group):
                    raise AssertionError('only member/assistant can query: %s' % msg)
        # 2. get group members
        members = facebook.members(identifier=group)
        if members is None or len(members) == 0:
            text = 'Group members not found: %s' % group
            return TextContent.new(text=text)
        # 3. respond group members for sender
        user = facebook.current_user
        assert user is not None, 'current user not set'
        if facebook.is_owner(member=user.identifier, group=group):
            return GroupCommand.reset(group=group, members=members)
        else:
            return GroupCommand.invite(group=group, members=members)


# register
GroupCommandProcessor.register(command=GroupCommand.QUERY, processor_class=QueryCommandProcessor)



# -*- coding: utf-8 -*-
#
#   DIM-SDK : Decentralized Instant Messaging Software Development Kit
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Forward Content Processor
    ~~~~~~~~~~~~~~~~~~~~~~~~~

"""

from typing import Optional

from dimp import ID
from dimp import InstantMessage
from dimp import ContentType, Content, ForwardContent

from .processor import ContentProcessor


#
#   Forward Content Processor
#
class ForwardContentProcessor(ContentProcessor):

    #
    #   main
    #
    def process(self, content: Content, sender: ID, msg: InstantMessage) -> Optional[Content]:
        assert isinstance(content, ForwardContent), 'forward content error: %s' % content
        r_msg = content.forward
        messenger = self.messenger
        # call messenger to process it
        r_msg = messenger.process_reliable(msg=r_msg)
        if r_msg is not None:
            messenger.send_message(msg=r_msg)
        # else:
        #     receiver = content.forward.envelope.receiver
        #     text = 'Message forwarded: %s' % receiver
        #     return ReceiptCommand.new(message=text)

        # NOTICE: decrypt failed, not for you?
        #         it means you are asked to re-pack and forward this message
        # TODO: override to catch the exception 'receiver error ...'
        return None


# register
ContentProcessor.register(content_type=ContentType.Forward, processor_class=ForwardContentProcessor)



# -*- coding: utf-8 -*-
#
#   DIM-SDK : Decentralized Instant Messaging Software Development Kit
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import ABC, abstractmethod
from typing import Optional, Any

"""
    Notification observer
    ~~~~~~~~~~~~~~~~~~~~~

    Notification object with name, sender and extra info
"""


class Notification:

    def __init__(self, name: str, sender: Any, info: dict=None):
        super().__init__()
        self.__name = name
        self.__sender = sender
        self.__info = info

    @property
    def name(self) -> str:
        return self.__name

    @property
    def sender(self) -> Any:
        return self.__sender

    @property
    def info(self) -> Optional[dict]:
        return self.__info

    @info.setter
    def info(self, value: dict):
        self.__info = value


class Observer(ABC):

    @abstractmethod
    def received_notification(self, notification: Notification):
        """
        Callback for notification

        :param notification: notification with name, sender and extra info
        :return:
        """
        pass



# -*- coding: utf-8 -*-
#
#   DIM-SDK : Decentralized Instant Messaging Software Development Kit
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Invite Group Command Processor
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    1. add new member(s) to the group
    2. any member or assistant can invite new member
"""

from typing import Optional

from dimp import ID
from dimp import InstantMessage
from dimp import Content
from dimp import GroupCommand, InviteCommand

from .history import GroupCommandProcessor


class InviteCommandProcessor(GroupCommandProcessor):

    def __is_reset(self, sender: ID, group: ID, invite_list: list) -> bool:
        """
        Check whether this is a 'reset' command

        :param sender:      owner ID?
        :param group:       group ID
        :param invite_list: new members list
        :return: True on owner invites owner
        """
        if self.contains_owner(members=invite_list, group=group):
            # NOTICE: owner invite owner?
            #         it's a Reset command!
            if self.facebook.is_owner(member=sender, group=group):
                return True

    def __reset(self, content: Content, sender: ID, msg: InstantMessage) -> Content:
        """
        Call reset command processor

        :param content: invite(reset) command
        :param sender:  owner ID?
        :param msg:     instant message
        :return: response from invite command processor
        """
        cpu: GroupCommandProcessor = self.cpu(command=GroupCommand.RESET)
        assert cpu is not None, 'failed to get "reset" command processor'
        return cpu.process(content=content, sender=sender, msg=msg)

    def __add(self, invite_list: list, group: ID) -> Optional[list]:
        facebook = self.facebook
        # existed members
        members: list = facebook.members(identifier=group)
        if members is None:
            members = []
        # added member(s)
        add_list = []
        for item in invite_list:
            if item in members:
                continue
            # new member found
            add_list.append(item)
            members.append(item)
        # response added-list after changed
        if len(add_list) > 0:
            if facebook.save_members(members=members, identifier=group):
                return add_list

    #
    #   main
    #
    def process(self, content: Content, sender: ID, msg: InstantMessage) -> Optional[Content]:
        assert isinstance(content, InviteCommand), 'group command error: %s' % content
        facebook = self.facebook
        group: ID = facebook.identifier(content.group)
        # 0. check whether group info empty
        if self.is_empty(group=group):
            # NOTICE:
            #     group membership lost?
            #     reset group members
            return self.__reset(content=content, sender=sender, msg=msg)
        # 1. check permission
        if not facebook.exists_member(member=sender, group=group):
            if not facebook.exists_assistant(member=sender, group=group):
                if not facebook.is_owner(member=sender, group=group):
                    raise AssertionError('only member/assistant can invite: %s' % msg)
        # 2. get inviting members
        invite_list: list = self.members(content=content)
        if invite_list is None or len(invite_list) == 0:
            raise ValueError('invite command error: %s' % content)
        # 2.1. check founder for reset command
        if self.__is_reset(sender=sender, group=group, invite_list=invite_list):
            # NOTICE:
            #     owner invites owner?
            #     it means this should be a 'reset' command
            return self.__reset(content=content, sender=sender, msg=msg)
        # 2.2. get invited-list
        add_list = self.__add(invite_list=invite_list, group=group)
        if add_list is not None:
            content['added'] = add_list
        # 3. response (no need to response this group command)
        return None


# register
GroupCommandProcessor.register(command=GroupCommand.INVITE, processor_class=InviteCommandProcessor)



# -*- coding: utf-8 -*-
#
#   DIM-SDK : Decentralized Instant Messaging Software Development Kit
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Session Server
    ~~~~~~~~~~~~~~

    for login user
"""

import numpy

from dimp import ID

from ..crypto import hex_encode


class Session:

    def __init__(self, identifier: ID, client_address):
        super().__init__()
        # user ID
        self.__identifier = identifier
        # (IP, port)
        self.__client_address = client_address
        # generate session key
        self.__session_key = hex_encode(bytes(numpy.random.bytes(32)))
        # valid flag: when handshake accepted, this should be set to True
        self.valid = False
        # active status: when the client entered background, it should be set to False
        self.active = True

    def __str__(self):
        clazz = self.__class__.__name__
        return '<%s:%s %s|%s valid=%d active=%d />' % (clazz,
                                                       self.__session_key,
                                                       self.__client_address,
                                                       self.__identifier,
                                                       self.valid,
                                                       self.active)

    @property
    def identifier(self) -> ID:
        return self.__identifier

    @property
    def client_address(self):  # (IP, port)
        return self.__client_address

    @property
    def session_key(self) -> str:
        return self.__session_key


class SessionServer:

    def __init__(self):
        super().__init__()
        # memory cache
        self.__pool = {}  # {identifier: [session]}

    def all(self, identifier: ID) -> list:
        """ Get all sessions of this user """
        return self.__pool.get(identifier)

    def clear(self, identifier: ID) -> bool:
        """ Remove all sessions of this user """
        if identifier in self.__pool:
            self.__pool.pop(identifier)
            return True

    def add(self, session: Session) -> bool:
        """ Add a session with ID into memory cache """
        identifier = session.identifier
        # 1. get all sessions with identifier
        array: list = self.all(identifier)
        if array is None:
            # 2.1. set a list contains this session object
            self.__pool[identifier] = [session]
            return True
        else:
            # 2.2. check each session with client address
            client_address = session.client_address
            for item in array:
                assert isinstance(item, Session), 'session error: %s' % item
                if item.client_address == client_address:
                    # already exists
                    return False
            # 3. add this session object to the current array
            array.append(session)
            return True

    def remove(self, session: Session) -> bool:
        """ Remove the session from memory cache """
        identifier = session.identifier
        # 1. get all sessions with identifier
        array: list = self.all(identifier)
        if array is not None:
            count = len(array)
            if count == 0:
                # 2.1. empty array, remove it
                self.clear(identifier)
            elif count == 1:
                # 2.2. check the only one session
                if array[0].client_address == session.client_address:
                    # exactly
                    self.clear(identifier)
                    return True
            else:
                # 2.3. check each session with client address
                for item in array:
                    assert isinstance(item, Session), 'session error: %s' % item
                    if item.client_address == session.client_address:
                        # got it
                        array.remove(session)
                        return True

    def get(self, identifier: ID, client_address) -> Session:
        """ Search session with ID and client address """
        array: list = self.all(identifier)
        if array is not None:
            for item in array:
                assert isinstance(item, Session), 'session error: %s' % item
                if item.client_address == client_address:
                    # got it
                    return item

    def new(self, identifier: ID, client_address):
        """ Session factory """
        session = self.get(identifier=identifier, client_address=client_address)
        if session is None:
            # create a new session
            session = Session(identifier=identifier, client_address=client_address)
            self.add(session=session)
        return session

    #
    #   Users
    #
    def all_users(self) -> list:
        """ Get all users """
        return list(self.__pool.keys())

    def online_users(self) -> list:
        """ Get online users """
        keys = self.all_users()
        array = []
        for identifier in keys:
            sessions = self.all(identifier)
            if sessions is None:
                # should not happen
                continue
            for item in sessions:
                assert isinstance(item, Session), 'session error: %s' % item
                if item.valid and item.active:
                    # got it
                    array.append(identifier)
                    break
        return array



# -*- coding: utf-8 -*-
#
#   DIM-SDK : Decentralized Instant Messaging Software Development Kit
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Chat Bot: XiaoI
    ~~~~~~~~~~~~~~~

    AI chat bot powered by XiaoI
"""

import urllib.request
from typing import Optional

import numpy

from .chatbot import ChatBot

from ..crypto import hex_encode, sha1


def sha_hex(string: str) -> str:
    return hex_encode(sha1(string.encode('utf-8')))


class XiaoI(ChatBot):

    def __init__(self, app_key: str, app_secret: str):
        super().__init__()
        self.user_id = 'dimchat'
        self.app_key = app_key
        self.app_secret = app_secret
        # self.api_url = 'http://nlp.xiaoi.com/ask.do'
        self.api_url = 'http://robot.open.xiaoi.com/ask.do'
        # parameters for auth
        self.realm = 'xiaoi.com'
        self.uri = '/ask.do'
        self.method = 'POST'
        # ignore responses
        self.ignores = ['默认回复', '重复回复']

    def __request(self, text: str) -> str:
        return 'type=0&platform=dim&userId=%s&question=%s' % (self.user_id, text)

    def __auth(self) -> str:
        random: bytes = numpy.random.bytes(40)
        nonce = hex_encode(random)
        # sign
        ha1 = sha_hex(self.app_key + ':' + self.realm + ':' + self.app_secret)
        ha2 = sha_hex(self.method + ':' + self.uri)
        sign = sha_hex(ha1 + ':' + nonce + ':' + ha2)
        return 'app_key="' + self.app_key + '", nonce="' + nonce + '", signature="' + sign + '"'

    def __post(self, text: str) -> str:
        request = self.__request(text=text)
        headers = {
            'content-type': 'application/x-www-form-urlencoded; charset=UTF-8',
            'X-Auth': self.__auth(),
        }
        http_post = urllib.request.Request(self.api_url, data=request.encode('utf-8'), headers=headers)
        response = urllib.request.urlopen(http_post)
        data: bytes = response.read()
        if data is not None:
            return data.decode('utf-8')

    def __fetch(self, response: str) -> Optional[str]:
        # check blah blah
        if response in self.ignores:
            # no answer, ignore it
            return None
        # got it
        return response

    def ask(self, question: str, user: str=None) -> str:
        if user is not None:
            self.user_id = user
        response = self.__post(text=question)
        if response is not None:
            return self.__fetch(response)



# -*- coding: utf-8 -*-
#
#   DIM-SDK : Decentralized Instant Messaging Software Development Kit
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Expel Group Command Processor
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    1. remove group member(s)
    2. only group owner or assistant can expel member
"""

from typing import Optional

from dimp import ID
from dimp import InstantMessage
from dimp import Content
from dimp import GroupCommand, ExpelCommand

from .history import GroupCommandProcessor


class ExpelCommandProcessor(GroupCommandProcessor):

    def __remove(self, expel_list: list, group: ID) -> Optional[list]:
        facebook = self.facebook
        # existed members
        members: list = facebook.members(identifier=group)
        if members is None:
            members = []
        # removed member(s)
        remove_list = []
        for item in expel_list:
            if item not in members:
                continue
            # expelled member found
            remove_list.append(item)
            members.remove(item)
        # response removed-list after changed
        if len(remove_list) > 0:
            if facebook.save_members(members=members, identifier=group):
                return remove_list

    #
    #   main
    #
    def process(self, content: Content, sender: ID, msg: InstantMessage) -> Optional[Content]:
        assert isinstance(content, ExpelCommand), 'group command error: %s' % content
        facebook = self.facebook
        group: ID = facebook.identifier(content.group)
        # 1. check permission
        if not facebook.is_owner(member=sender, group=group):
            if not facebook.exists_assistant(member=sender, group=group):
                raise AssertionError('only owner/assistant can expel: %s' % msg)
        # 2.1. get expelling members
        expel_list: list = self.members(content=content)
        if expel_list is None or len(expel_list) == 0:
            raise ValueError('expel command error: %s' % content)
        # 2.2. get removed-list
        remove_list = self.__remove(expel_list=expel_list, group=group)
        if remove_list is not None:
            content['removed'] = remove_list
        # 3. response (no need to response this group command)
        return None


# register
GroupCommandProcessor.register(command=GroupCommand.EXPEL, processor_class=ExpelCommandProcessor)



# -*- coding: utf-8 -*-
#
#   DIM-SDK : Decentralized Instant Messaging Software Development Kit
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Command Processor
    ~~~~~~~~~~~~~~~~~

"""

from typing import Optional

from dimp import ID
from dimp import InstantMessage
from dimp import ContentType, Content, TextContent
from dimp import Command

from .processor import ContentProcessor


class CommandProcessor(ContentProcessor):

    UNKNOWN = "unknown"

    def __init__(self, messenger):
        super().__init__(messenger=messenger)
        self.__command_processors: dict = {}

    #
    #   Runtime
    #
    __command_processor_classes = {}  # class map

    @classmethod
    def register(cls, command: str, processor_class=None) -> bool:
        if processor_class is None:
            cls.__command_processor_classes.pop(command, None)
        elif processor_class == CommandProcessor:
            raise TypeError('should not add CommandProcessor itself!')
        else:
            assert issubclass(processor_class, CommandProcessor), 'cpu class error' % processor_class
            cls.__command_processor_classes[command] = processor_class
        return True

    def cpu(self, command: str):
        # 1. get from pool
        processor = self.__command_processors.get(command)
        if processor is not None:
            return processor
        # 2. get CPU class by command name
        clazz = self.__command_processor_classes.get(command)
        if clazz is None:
            if command == self.UNKNOWN:
                raise LookupError('default CPU not set yet')
            # call default CPU
            return self.cpu(command=self.UNKNOWN)
        # 3. create CPU with messenger
        processor = self._create_processor(clazz)
        assert processor is not None, 'failed to create CPU for command: %s' % command
        self.__command_processors[command] = processor
        return processor

    #
    #   main
    #
    def process(self, content: Content, sender: ID, msg: InstantMessage) -> Optional[Content]:
        assert type(self) == CommandProcessor, 'override me!'
        assert isinstance(content, Command), 'command error: %s' % content
        # process command by name
        cpu = self.cpu(command=content.command)
        assert cpu is not self, 'Dead cycle! command: %s' % content
        return cpu.process(content=content, sender=sender, msg=msg)


#
#   Default Command Processor
#
class _DefaultCommandProcessor(CommandProcessor):

    #
    #   main
    #
    def process(self, content: Content, sender: ID, msg: InstantMessage) -> Content:
        assert isinstance(content, Command), 'command error: %s' % content
        res = TextContent.new(text='Command (name: %s) not support yet!' % content.command)
        # check group message
        group = content.group
        if group is not None:
            res.group = group
        return res


# register
ContentProcessor.register(content_type=ContentType.Command, processor_class=CommandProcessor)
CommandProcessor.register(command=CommandProcessor.UNKNOWN, processor_class=_DefaultCommandProcessor)



# -*- coding: utf-8 -*-
#
#   DIM-SDK : Decentralized Instant Messaging Software Development Kit
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Content Processor
    ~~~~~~~~~~~~~~~~~

"""

import weakref
from typing import Optional, Union

from dimp import ID
from dimp import InstantMessage
from dimp import ContentType, Content, TextContent


class ContentProcessor:

    def __init__(self, messenger):
        super().__init__()
        self.__content_processors: dict = {}
        self.__messenger = weakref.ref(messenger)

    @property
    def messenger(self):  # Messenger
        return self.__messenger()

    @property
    def facebook(self):  # Facebook
        return self.messenger.facebook

    def get_context(self, key: str):
        return self.messenger.get_context(key)

    def set_context(self, key: str, value):
        return self.messenger.set_context(key=key, value=value)

    #
    #   Runtime
    #
    def _create_processor(self, clazz):
        assert issubclass(clazz, ContentProcessor), 'cpu class error: %s' % clazz
        return clazz(self.messenger)

    __content_processor_classes = {}  # class map

    @classmethod
    def register(cls, content_type: Union[ContentType, int], processor_class=None) -> bool:
        if isinstance(content_type, ContentType):
            content_type = content_type.value
        if processor_class is None:
            cls.__content_processor_classes.pop(content_type, None)
        elif processor_class == ContentProcessor:
            raise TypeError('should not add ContentProcessor itself!')
        else:
            assert issubclass(processor_class, ContentProcessor), 'cpu class error: %s' % processor_class
            cls.__content_processor_classes[content_type] = processor_class
        return True

    def cpu(self, content_type: Union[ContentType, int]):
        if isinstance(content_type, ContentType):
            content_type = content_type.value
        # 1. get from pool
        processor = self.__content_processors.get(content_type)
        if processor is not None:
            return processor
        # 2. get CPU class by content type
        clazz = self.__content_processor_classes.get(content_type)
        if clazz is None:
            if ContentType.Unknown == content_type:
                raise LookupError('default CPU not register yet')
            # call default CPU
            return self.cpu(content_type=ContentType.Unknown)
        # 3. create CPU with messenger
        processor = self._create_processor(clazz)
        assert processor is not None, 'failed to create CPU for content type: %s' % content_type
        self.__content_processors[content_type] = processor
        return processor

    #
    #   main
    #
    def process(self, content: Content, sender: ID, msg: InstantMessage) -> Optional[Content]:
        assert type(self) == ContentProcessor, 'override me!'
        # process content by type
        cpu = self.cpu(content_type=content.type)
        assert cpu is not self, 'Dead cycle! content: %s' % content
        return cpu.process(content=content, sender=sender, msg=msg)


#
#   Default Content Processor
#
class _DefaultContentProcessor(ContentProcessor):

    #
    #   main
    #
    def process(self, content: Content, sender: ID, msg: InstantMessage) -> Content:
        res = TextContent.new(text='Content (type: %s) not support yet!' % content.type)
        # check group message
        group = content.group
        if group is not None:
            res.group = group
        return res


# register
ContentProcessor.register(content_type=ContentType.Unknown, processor_class=_DefaultContentProcessor)



# -*- coding: utf-8 -*-
#
#   DIM-SDK : Decentralized Instant Messaging Software Development Kit
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Station & Service Provider (SP)
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Station : DIM network server node
    SP      : DIM network service provider
"""

from dimp import PublicKey
from dimp import ID, NetworkID, User, Group

from .certificate import CertificateAuthority


class Station(User):

    def __init__(self, identifier: ID, host: str=None, port: int=0):
        super().__init__(identifier=identifier)
        assert identifier.type == NetworkID.Station, 'Station ID type error: %s' % identifier
        self.__host = host
        self.__port = port

    def __str__(self):
        clazz = self.__class__.__name__
        identifier = self.identifier
        network = identifier.address.network
        number = identifier.address.number
        name = self.name
        host = self.host
        port = self.port
        return '<%s: %s(%s|%d) "%s" host="%s" port=%d />' % (clazz, identifier, network, number, name, host, port)

    @property
    def host(self) -> str:
        if self.__host is None:
            profile = self.profile
            if profile is not None:
                value = profile.get_property('host')
                if value is not None:
                    self.__host = str(value)
            if self.__host is None:
                self.__host = '0.0.0.0'
        return self.__host

    @property
    def port(self) -> int:
        if self.__port is 0:
            profile = self.profile
            if profile is not None:
                value = profile.get_property('port')
                if value is not None:
                    self.__port = int(value)
            if self.__port is 0:
                self.__port = 9394
        return self.__port

    @classmethod
    def new(cls, station: dict):
        # ID
        identifier = ID(station['ID'])
        # host
        host = station['host']
        # port
        port = station.get('port')
        if port is None:
            port = 9394
        else:
            port = int(port)
        # Certificate Authority
        if 'CA' in station['CA']:
            certificate = CertificateAuthority(station['CA'])
        else:
            certificate = None
        # Service Provider's ID
        if 'SP' in station['SP']:
            provider = station['SP']
            if isinstance(provider, dict):
                provider = ID(provider['ID'])
            elif isinstance(provider, str):
                provider = ID(provider)
            else:
                raise TypeError('Service Provider data format error')
        else:
            provider = None

        # create Station object
        if identifier.type == NetworkID.Station:
            self = Station(identifier=identifier, host=host, port=port)
            self.provider = provider
            self.certificate = certificate
            return self
        else:
            raise ValueError('Station ID error')


class ServiceProvider(Group):

    def __init__(self, identifier: ID):
        super().__init__(identifier=identifier)
        assert identifier.type == NetworkID.Provider, 'Service Provider ID type error: %s' % identifier

    @property
    def stations(self):
        return self.members

    @classmethod
    def new(cls, provider: dict):
        # ID
        identifier = ID(provider['ID'])
        # Public Key
        if 'publicKey' in provider:
            public_key = PublicKey(provider['publicKey'])
        elif 'PK' in provider:
            public_key = PublicKey(provider['PK'])
        else:
            public_key = None
        # Certificate Authority
        if 'CA' in provider['CA']:
            certificate = CertificateAuthority(provider['CA'])
            if public_key is None:
                # get public key from the CA info
                public_key = certificate.info.key
        else:
            certificate = None

        # create ServiceProvider object
        if identifier.type == NetworkID.Provider:
            self = ServiceProvider(identifier=identifier)
            self.publicKey = public_key
            self.certificate = certificate
            return self
        else:
            raise ValueError('Service provider ID error')



# -*- coding: utf-8 -*-
#
#   DIM-SDK : Decentralized Instant Messaging Software Development Kit
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Block Protocol
    ~~~~~~~~~~~~~~

    Ignore all messages in this conversation, which ID(user/group) contains in 'list'.
    If value of 'list' is None, means querying block-list from station
"""

from typing import Optional

from dimp import Command


class BlockCommand(Command):
    """
        Block Command
        ~~~~~~~~~~~~~

        data format: {
            type : 0x88,
            sn   : 123,

            command : "block", // command name
            list    : []       // block-list
        }
    """

    BLOCK = 'block'

    def __new__(cls, cmd: dict):
        """
        Create block command

        :param cmd: command info
        :return: BlockCommand object
        """
        if cmd is None:
            return None
        elif cls is BlockCommand:
            if isinstance(cmd, BlockCommand):
                # return BlockCommand object directly
                return cmd
        # new BlockCommand(dict)
        return super().__new__(cls, cmd)

    def __init__(self, content: dict):
        if self is content:
            # no need to init again
            return
        super().__init__(content)

    #
    #   block-list
    #
    @property
    def block_list(self) -> Optional[list]:
        return self.get('list')

    @block_list.setter
    def block_list(self, value: list):
        if value is None:
            self.pop('list', None)
        else:
            self['list'] = value

    #
    #   Factories
    #
    @classmethod
    def new(cls, content: dict=None, block: list=None):
        """
        Create block command

        :param content: command info
        :param block: block-list
        :return: BlockCommand object
        """
        if content is None:
            # create empty content
            content = {}
        # set block-list
        if block is not None:
            content['list'] = block
        # new BlockCommand(dict)
        return super().new(content=content, command=cls.BLOCK)


# register command class
Command.register(command=BlockCommand.BLOCK, command_class=BlockCommand)



# -*- coding: utf-8 -*-
#
#   DIM-SDK : Decentralized Instant Messaging Software Development Kit
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Chat Bot: Tuling
    ~~~~~~~~~~~~~~~~

    AI chat bot powered by Tuling
"""

import json
import urllib.request
from typing import Optional

from .chatbot import ChatBot


class Tuling(ChatBot):

    def __init__(self, api_key: str):
        super().__init__()
        self.user_id = 'dimchat'
        self.api_key = api_key
        self.api_url = 'http://openapi.tuling123.com/openapi/api/v2'
        self.province = 'Guangdong'
        self.city = 'Guangzhou'
        # ignore codes
        self.ignores = [4003]

    def __request(self, text: str) -> str:
        return json.dumps({
            'perception': {
                'inputText': {
                    'text': text,
                },
            },
            'selfInfo': {
                'location': {
                    'province': self.province,
                    'city': self.city,
                }
            },
            'userInfo': {
                'apiKey': self.api_key,
                'userId': self.user_id,
            }
        })

    def __post(self, text: str) -> dict:
        request = self.__request(text=text)
        headers = {'content-type': 'application/json'}
        http_post = urllib.request.Request(self.api_url, data=request.encode('utf-8'), headers=headers)
        response = urllib.request.urlopen(http_post)
        data: bytes = response.read()
        if data is not None and len(data) > 0:
            # assert data[0] == ord('{') and data[-1] == ord('}'), 'response error: %s' % data
            return json.loads(data)

    def __fetch(self, response: dict) -> Optional[str]:
        # get code
        intent: dict = response.get('intent')
        if intent is not None:
            code = intent.get('code')
            if code in self.ignores:
                # requests limited for test, ignore it
                return None
        # get text
        results: list = response.get('results')
        if results is not None and len(results) > 0:
            values: dict = results[0].get('values')
            if values is not None:
                return values.get('text')

    def ask(self, question: str, user: str=None) -> str:
        if user is not None:
            self.user_id = user
        response = self.__post(text=question)
        if response is not None:
            return self.__fetch(response)



# -*- coding: utf-8 -*-
#
#   DIM-SDK : Decentralized Instant Messaging Software Development Kit
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Key Store
    ~~~~~~~~~

    Memory cache for reused passwords (symmetric key)
"""

import os
from typing import Optional

from mkm.dos import JSONFile

from dimp import User
from dimp import KeyCache


class KeyStore(KeyCache):

    def __init__(self):
        super().__init__()
        self.__user: User = None
        self.__base_dir: str = '/tmp/.dim/'

    @property
    def user(self) -> Optional[User]:
        return self.__user

    @user.setter
    def user(self, value: User):
        if self.__user is not None:
            # save key map for old user
            self.flush()
            if self.__user == value:
                # user not changed
                return
        if value is None:
            self.__user = None
            return
        # change current user
        self.__user = value
        keys = self.load_keys()
        if keys is None:
            # failed to load cached keys for new user
            return
        # update key map
        self.update_keys(key_map=keys)

    @property
    def directory(self) -> str:
        return self.__base_dir

    @directory.setter
    def directory(self, value: str):
        self.__base_dir = value

    # '/tmp/.dim/protected/{ADDRESS}/keystore.js'
    def __path(self) -> Optional[str]:
        if self.__user is None:
            return None
        return os.path.join(self.__base_dir, 'protected', self.__user.identifier, 'keystore.js')

    def save_keys(self, key_map: dict) -> bool:
        # write key table to persistent storage
        path = self.__path()
        if path is None:
            return False
        return JSONFile(path).write(key_map)

    def load_keys(self) -> Optional[dict]:
        # load key table from persistent storage
        path = self.__path()
        if path is None:
            return None
        return JSONFile(path).read()



# -*- coding: utf-8 -*-
#
#   DIM-SDK : Decentralized Instant Messaging Software Development Kit
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Group History Processors
    ~~~~~~~~~~~~~~~~~~~~~~~~

"""

from typing import Optional

from dimp import ID
from dimp import InstantMessage
from dimp import ContentType, Content
from dimp import Command, GroupCommand

from .processor import ContentProcessor
from .command import CommandProcessor


class HistoryCommandProcessor(CommandProcessor):

    def __init__(self, messenger):
        super().__init__(messenger=messenger)
        # lazy
        self.__gpu: GroupCommandProcessor = None

    @property
    def gpu(self):  # GroupCommandProcessor
        if self.__gpu is None:
            self.__gpu = self._create_processor(GroupCommandProcessor)
        return self.__gpu

    #
    #   main
    #
    def process(self, content: Content, sender: ID, msg: InstantMessage) -> Optional[Content]:
        assert type(self) == HistoryCommandProcessor, 'override me!'
        assert isinstance(content, Command), 'history cmd error: %s' % content
        if content.group is None:
            # get command processor
            cpu = self.cpu(command=content.command)
            # if cpu is None:
            #     return TextContent.new(text='History command (name: %s) not support yet!' % content.command)
        else:
            # get group command processor
            cpu = self.gpu
        assert cpu is not self, 'Dead cycle! history cmd: %s' % content
        return cpu.process(content=content, sender=sender, msg=msg)


class GroupCommandProcessor(HistoryCommandProcessor):

    def members(self, content: GroupCommand) -> Optional[list]:
        array = content.members
        if array is None:
            item = content.member
            if item is None:
                return None
            array = [item]
        return self.convert_members(array)

    def convert_members(self, array: list) -> list:
        facebook = self.facebook
        results = []
        for item in array:
            identifier = facebook.identifier(item)
            if identifier is None:
                raise ValueError('Member ID error: %s' % item)
            results.append(identifier)
        return results

    def contains_owner(self, members: list, group: ID) -> bool:
        facebook = self.facebook
        for item in members:
            user = facebook.identifier(item)
            if facebook.is_owner(member=user, group=group):
                return True

    def is_empty(self, group: ID) -> bool:
        """
        Check whether group info empty (lost)

        :param group: group ID
        :return: True on members, owner not found
        """
        facebook = self.facebook
        members = facebook.members(identifier=group)
        if members is None or len(members) == 0:
            return True
        owner = facebook.owner(identifier=group)
        if owner is None:
            return True

    #
    #   main
    #
    def process(self, content: Content, sender: ID, msg: InstantMessage) -> Optional[Content]:
        assert type(self) == GroupCommandProcessor, 'override me!'
        assert isinstance(content, Command), 'group cmd error: %s' % content
        # process command by name
        cpu = self.cpu(command=content.command)
        # if cpu is None:
        #     return TextContent.new(text='Group command (name: %s) not support yet!' % content.command)
        assert cpu is not self, 'Dead cycle! group cmd: %s' % content
        return cpu.process(content=content, sender=sender, msg=msg)


# register
ContentProcessor.register(content_type=ContentType.History, processor_class=HistoryCommandProcessor)



# -*- coding: utf-8 -*-
#
#   DIM-SDK : Decentralized Instant Messaging Software Development Kit
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Facebook
    ~~~~~~~~

    Barrack for cache entities
"""

import time
import weakref
from abc import abstractmethod
from typing import Optional

from dimp import PrivateKey, SignKey, DecryptKey
from dimp import NetworkID, ID, Address
from dimp import User, Group
from dimp import Meta, Profile
from dimp import Barrack

from .ans import AddressNameService
from .network import ServiceProvider, Station, Robot
from .group import Polylogue


class Facebook(Barrack):

    EXPIRES = 3600  # profile expires (1 hour)

    def __init__(self):
        super().__init__()
        self.__ans: weakref.ReferenceType = None
        # caches
        self.__profiles: dict = {}
        self.__private_keys: dict = {}
        self.__contacts: dict = {}
        self.__members: dict = {}

    @property
    def ans(self) -> AddressNameService:
        if self.__ans is not None:
            return self.__ans()

    @ans.setter
    def ans(self, value: AddressNameService):
        self.__ans = weakref.ref(value)

    def ans_get(self, name: str) -> ID:
        ans = self.ans
        if ans is not None:
            return ans.identifier(name=name)

    #
    #   Meta
    #
    @staticmethod
    def verify_meta(meta: Meta, identifier: ID) -> bool:
        assert meta is not None, 'meta should not be empty'
        return meta.match_identifier(identifier)

    def cache_meta(self, meta: Meta, identifier: ID) -> bool:
        if not self.verify_meta(meta=meta, identifier=identifier):
            return False
        return super().cache_meta(meta=meta, identifier=identifier)

    @abstractmethod
    def save_meta(self, meta: Meta, identifier: ID) -> bool:
        """ Save meta into local storage """
        raise NotImplemented

    @abstractmethod
    def load_meta(self, identifier: ID) -> Optional[Meta]:
        """ Load meta from local storage """
        raise NotImplemented

    #
    #   Profile
    #
    def verify_profile(self, profile: Profile, identifier: ID=None) -> bool:
        assert profile is not None, 'profile should not be empty'
        if identifier is not None and profile.identifier != identifier:
            # profile ID not match
            return False
        if identifier is None:
            identifier = self.identifier(profile.identifier)
            assert identifier is not None, 'profile error: %s ' % profile
        # NOTICE: if this is a group profile,
        #             verify it with each member's meta.key
        #         else (this is a user profile)
        #             verify it with the user's meta.key
        if identifier.is_group:
            # check by each member
            members = self.members(identifier=identifier)
            if members is not None:
                for item in members:
                    meta = self.meta(identifier=item)
                    if meta is None:
                        # FIXME: meta not found for this member
                        continue
                    if profile.verify(public_key=meta.key):
                        return True
            # DISCUSS: what to do about assistants?

            # check by owner
            owner = self.owner(identifier=identifier)
            if owner is None:
                if identifier.type == NetworkID.Polylogue:
                    # NOTICE: if this is a polylogue profile
                    #             verify it with the founder's meta.key
                    #             (which equals to the group's meta.key)
                    meta = self.meta(identifier=identifier)
                else:
                    # FIXME: owner not found for this group
                    return False
            elif owner in members:
                # already checked
                return False
            else:
                meta = self.meta(identifier=owner)
        else:
            assert identifier.is_user, 'profile ID error: %s' % identifier
            meta = self.meta(identifier=identifier)
        if meta is not None:
            return profile.verify(public_key=meta.key)

    def cache_profile(self, profile: Profile, identifier: ID=None) -> bool:
        if profile is None:
            # remove from cache if exists
            self.__profiles.pop(identifier, None)
            return False
        if not self.verify_profile(profile=profile, identifier=identifier):
            return False
        if identifier is None:
            identifier = self.identifier(string=profile.identifier)
        self.__profiles[identifier] = profile
        return True

    @abstractmethod
    def save_profile(self, profile: Profile, identifier: ID=None) -> bool:
        """ Save profile into database """
        raise NotImplemented

    @abstractmethod
    def load_profile(self, identifier: ID) -> Optional[Profile]:
        """ Load profile from database """
        raise NotImplemented

    #
    #   Private keys
    #
    def verify_private_key(self, key: PrivateKey, identifier: ID) -> bool:
        assert key is not None, 'private key should not be empty'
        meta = self.meta(identifier=identifier)
        assert meta is not None, 'meta not found: %s' % identifier
        assert meta.key is not None, 'meta error: %s, %s' % (identifier, meta)
        return meta.key.match(private_key=key)

    def cache_private_key(self, key: PrivateKey, identifier: ID) -> bool:
        assert identifier.is_user, 'user ID error: %s' % identifier
        if key is None:
            self.__private_keys.pop(identifier, None)
            return False
        if not self.verify_private_key(key=key, identifier=identifier):
            return False
        self.__private_keys[identifier] = key
        return True

    @abstractmethod
    def save_private_key(self, key: PrivateKey, identifier: ID) -> bool:
        """ Save private key into safety storage """
        raise NotImplemented

    @abstractmethod
    def load_private_key(self, identifier: ID) -> Optional[PrivateKey]:
        """ Load private key from safety storage """
        raise NotImplemented

    #
    #   User Contacts
    #
    def cache_contacts(self, contacts: list, identifier: ID) -> bool:
        assert identifier.is_user, 'user ID error: %s' % identifier
        if contacts is None:
            self.__contacts.pop(identifier, None)
            return False
        self.__contacts[identifier] = contacts
        return True

    @abstractmethod
    def save_contacts(self, contacts: list, identifier: ID) -> bool:
        """ Save contacts into database """
        raise NotImplemented

    @abstractmethod
    def load_contacts(self, identifier: ID) -> Optional[list]:
        """ Load contacts from database """
        raise NotImplemented

    #
    #   Group Members
    #
    def cache_members(self, members: list, identifier: ID) -> bool:
        assert identifier.is_group, 'group ID error: %s' % identifier
        if members is None:
            self.__members.pop(identifier, None)
            return False
        self.__members[identifier] = members
        return True

    @abstractmethod
    def save_members(self, members: list, identifier: ID) -> bool:
        """ Save members into database """
        raise NotImplemented

    @abstractmethod
    def load_members(self, identifier: ID) -> Optional[list]:
        """ Load members from database """
        raise NotImplemented

    #
    #   All local users (for decrypting received message)
    #
    @property
    def local_users(self) -> Optional[list]:
        raise NotImplemented

    #
    #   Current user (for signing and sending message)
    #
    @property
    def current_user(self) -> Optional[User]:
        users = self.local_users
        if users is not None and len(users) > 0:
            return users[0]

    def __identifier(self, address: Address) -> Optional[ID]:
        """ generate ID from meta with address """
        identifier = ID.new(address=address)
        meta = self.meta(identifier=identifier)
        if meta is None:
            # failed to get meta for this address
            return None
        seed = meta.seed
        if seed is None or len(seed) == 0:
            return identifier
        identifier = meta.generate_identifier(address.network)
        self.cache_id(identifier)
        return identifier

    def create_identifier(self, string: str) -> ID:
        if isinstance(string, Address):
            # convert Address to ID
            return self.__identifier(address=string)
        # try from ANS record
        identifier = self.ans_get(name=string)
        if identifier is not None:
            return identifier
        # create by barrack
        return super().create_identifier(string=string)

    def create_user(self, identifier: ID) -> User:
        assert identifier.is_user, 'user ID error: %s' % identifier
        if identifier.is_broadcast:
            # create user 'anyone@anywhere'
            return User(identifier=identifier)
        # make sure meta exists
        assert self.meta(identifier) is not None, 'failed to get meta for user: %s' % identifier
        # TODO: check user type
        u_type = identifier.type
        if u_type == NetworkID.Main or u_type == NetworkID.BTCMain:
            return User(identifier=identifier)
        if u_type == NetworkID.Robot:
            return Robot(identifier=identifier)
        if u_type == NetworkID.Station:
            return Station(identifier=identifier)
        raise TypeError('unsupported user type: %s' % u_type)

    def create_group(self, identifier: ID) -> Group:
        assert identifier.is_group, 'group ID error: %s' % identifier
        if identifier.is_broadcast:
            # create group 'everyone@everywhere'
            return Group(identifier=identifier)
        # make sure meta exists
        assert self.meta(identifier) is not None, 'failed to get meta for group: %s' % identifier
        # TODO: check group type
        g_type = identifier.type
        if g_type == NetworkID.Polylogue:
            return Polylogue(identifier=identifier)
        if g_type == NetworkID.Chatroom:
            raise NotImplementedError('Chatroom not implemented')
        if g_type == NetworkID.Provider:
            return ServiceProvider(identifier=identifier)
        raise TypeError('unsupported group type: %s' % g_type)

    #
    #   EntityDataSource
    #
    def meta(self, identifier: ID) -> Optional[Meta]:
        info = super().meta(identifier=identifier)
        if info is not None:
            return info
        # load from local storage
        info = self.load_meta(identifier=identifier)
        if info is None:
            return None
        # no need to verify meta from local storage
        super().cache_meta(meta=info, identifier=identifier)
        return info

    EXPIRES_KEY = 'expires'

    def profile(self, identifier: ID) -> Optional[Profile]:
        info = self.__profiles.get(identifier)
        if info is not None:
            # check expired time
            timestamp = time.time() + self.EXPIRES
            expires = info.get(self.EXPIRES_KEY)
            if expires is None:
                # set expired time
                info[self.EXPIRES_KEY] = timestamp
                return info
            elif expires < timestamp:
                # not expired yet
                return info
        # load from local storage
        info = self.load_profile(identifier=identifier)
        if info is None:
            info = Profile.new(identifier=identifier)
        else:
            info.pop(self.EXPIRES_KEY, None)
        # no need to verify profile from local storage
        self.__profiles[identifier] = info
        return info

    #
    #   UserDataSource
    #
    def contacts(self, identifier: ID) -> Optional[list]:
        array = self.__contacts.get(identifier)
        if array is not None:
            return array
        # load from local storage
        array = self.load_contacts(identifier=identifier)
        if array is None:
            return None
        self.cache_contacts(contacts=array, identifier=identifier)
        return array

    def private_key_for_signature(self, identifier: ID) -> Optional[SignKey]:
        key = self.__private_keys.get(identifier)
        if key is not None:
            return key
        # load from local storage
        key = self.load_private_key(identifier=identifier)
        if key is None:
            return None
        self.cache_private_key(key=key, identifier=identifier)
        return key

    def private_keys_for_decryption(self, identifier: ID) -> Optional[list]:
        # DIMP v1.0:
        #     decrypt key not found, use the same with sign key?
        key = self.private_key_for_signature(identifier)
        if isinstance(key, DecryptKey):
            # TODO: support profile.key
            return [key]

    #
    #   GroupDataSource
    #
    def founder(self, identifier: ID) -> Optional[ID]:
        uid = super().founder(identifier=identifier)
        if uid is not None:
            return uid
        # check each member's public key with group meta
        members = self.members(identifier=identifier)
        if members is not None:
            meta = self.meta(identifier=identifier)
            if meta is not None:
                # if the member's public key matches with the group's meta,
                # it means this meta was generate by the member's private key
                for item in members:
                    m = self.meta(identifier=self.identifier(item))
                    if m is not None and meta.match_public_key(m.key):
                        # got it
                        return item
        # TODO: load founder from database

    def owner(self, identifier: ID) -> ID:
        uid = super().owner(identifier=identifier)
        if uid is not None:
            return uid
        # check group type
        if identifier.type == NetworkID.Polylogue:
            # Polylogue's owner is its founder
            return self.founder(identifier=identifier)
        # TODO: load owner from database

    def members(self, identifier: ID) -> Optional[list]:
        array = super().members(identifier=identifier)
        if array is None:
            # get from cache
            array = self.__members.get(identifier)
        if array is not None:
            return array
        # load from local storage
        array = self.load_members(identifier=identifier)
        if array is None:
            return None
        self.cache_members(members=array, identifier=identifier)
        return array

    def is_founder(self, member: ID, group: ID) -> bool:
        # check member's public key with group's meta.key
        g_meta = self.meta(identifier=group)
        if g_meta is None:
            raise LookupError('failed to get meta for group: %s' % group)
        m_meta = self.meta(identifier=member)
        if m_meta is None:
            raise LookupError('failed to get meta for member: %s' % member)
        return g_meta.match_public_key(m_meta.key)

    def is_owner(self, member: ID, group: ID) -> bool:
        if group.type == NetworkID.Polylogue:
            return self.is_founder(member=member, group=group)
        else:
            raise NotImplementedError('only Polylogue so far')

    def exists_member(self, member: ID, group: ID) -> bool:
        members = self.members(identifier=group)
        if member in members:
            return True
        owner = self.owner(identifier=group)
        if member == owner:
            return True

    #
    #   Group Assistants
    #
    def assistants(self, identifier: ID) -> Optional[list]:
        assert identifier.is_group, 'Group ID error: %s' % identifier
        identifier = self.ans_get(name='assistant')
        if identifier is not None:
            return [identifier]

    def exists_assistant(self, member: ID, group: ID) -> bool:
        assistants = self.assistants(identifier=group)
        if assistants is not None:
            return member in assistants



# -*- coding: utf-8 -*-
#
#   DIM-SDK : Decentralized Instant Messaging Software Development Kit
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Quit Group Command Processor
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    1. quit the group
    2. owner and assistant cannot quit
    3. assistant can be hired/fired by owner
"""

from typing import Optional

from dimp import ID
from dimp import InstantMessage
from dimp import Content
from dimp import GroupCommand, QuitCommand

from .history import GroupCommandProcessor


class QuitCommandProcessor(GroupCommandProcessor):

    def __remove(self, sender: ID, group: ID) -> bool:
        facebook = self.facebook
        members = facebook.members(identifier=group)
        if members is None:
            return False
        if sender not in members:
            return False
        members.remove(sender)
        return facebook.save_members(members=members, identifier=group)

    #
    #   main
    #
    def process(self, content: Content, sender: ID, msg: InstantMessage) -> Optional[Content]:
        assert isinstance(content, QuitCommand), 'group command error: %s' % content
        facebook = self.facebook
        group: ID = facebook.identifier(content.group)
        # 1. check permission
        if facebook.is_owner(member=sender, group=group):
            raise AssertionError('owner cannot quit: %s' % msg)
        if facebook.exists_assistant(member=sender, group=group):
            raise AssertionError('assistant cannot quit: %s' % msg)
        # 2. remove sender from group members
        self.__remove(sender=sender, group=group)
        # 3. response (no need to response this group command)
        return None


# register
GroupCommandProcessor.register(command=GroupCommand.QUIT, processor_class=QuitCommandProcessor)



# -*- coding: utf-8 -*-
#
#   DIM-SDK : Decentralized Instant Messaging Software Development Kit
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Receipt Protocol
    ~~~~~~~~~~~~~~~~

    As receipt returned to sender to proofing the message's received
"""

from typing import Optional

from dimp import Envelope
from dimp import Command


class ReceiptCommand(Command):
    """
        Receipt Command
        ~~~~~~~~~~~~~~~

        data format: {
            type : 0x88,
            sn   : 123,

            command  : "receipt", // command name
            message  : "...",
            //-- extra info
            sender   : "...",
            receiver : "...",
            time     : 0
        }
    """

    def __new__(cls, cmd: dict):
        """
        Create receipt command

        :param cmd: command info
        :return: ReceiptCommand object
        """
        if cmd is None:
            return None
        elif cls is ReceiptCommand:
            if isinstance(cmd, ReceiptCommand):
                # return ReceiptCommand object directly
                return cmd
        # new ReceiptCommand(dict)
        return super().__new__(cls, cmd)

    def __init__(self, content: dict):
        if self is content:
            # no need to init again
            return
        super().__init__(content)
        # lazy
        self.__envelope: Envelope = None

    # -------- setters/getters

    @property
    def message(self) -> Optional[str]:
        return self.get('message')

    @property
    def envelope(self) -> Optional[Envelope]:
        if self.__envelope is None:
            # envelope: { sender: "...", receiver: "...", time: 0 }
            env = self.get('envelope')
            if env is None and 'sender' in self and 'receiver' in self:
                env = self
            self.__envelope = Envelope(env)
        return self.__envelope

    #
    #   Factories
    #
    @classmethod
    def new(cls, content: dict=None, message: str=None, sn: int=0, envelope: Envelope=None):
        """
        Create receipt command

        :param content:  command info
        :param message:  receipt message
        :param sn:       serial number of the message responding to
        :param envelope: envelope of the message responding to
        :return: ReceiptCommand object
        """
        if content is None:
            # create empty content
            content = {}
        # set receipt message
        if message is not None:
            content['message'] = message
        # set sn for the message responding to
        if sn > 0:
            content['sn'] = sn
        # set envelope for the message responding to
        if envelope is not None:
            content['sender'] = envelope.sender
            content['receiver'] = envelope.receiver
            content['time'] = envelope.time
        # new ReceiptCommand(dict)
        return super().new(content=content, command=Command.RECEIPT)


# register command class
Command.register(command=Command.RECEIPT, command_class=ReceiptCommand)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
    Decentralized Instant Messaging Software Development Kit
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This is a new protocol designed for instant messaging (IM).
    The software provides accounts(user identity recognition) and
    communications between accounts safely by end-to-end encryption.
"""

from setuptools import setup, find_packages

__version__ = '0.3.7'
__author__ = 'Albert Moky'
__contact__ = 'albert.moky@gmail.com'

with open('README.md', 'r') as fh:
    readme = fh.read()

setup(
    name='dimsdk',
    version=__version__,
    url='https://github.com/dimchat/sdk-py',
    license='MIT',
    author=__author__,
    author_email=__contact__,
    description='Decentralized Instant Messaging SDK',
    long_description=readme,
    long_description_content_type='text/markdown',
    packages=find_packages(),
    package_data={
        '': ['*.js']
    },
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    install_requires=[
        'dimp>=0.8.9',
        # 'dkd>=0.6.9',
        'mkm>=0.8.10',
        'apns2',  # 0.4.1
        'numpy',  # 1.15.4
    ]
)



# -*- coding: utf-8 -*-
#
#   DIM-SDK : Decentralized Instant Messaging Software Development Kit
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Reset Group Command Processor
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    1. reset group members
    2. only group owner or assistant can reset group members

    3. specially, if the group members info lost,
       means you may not known who's the group owner immediately (and he may be not online),
       so we accept the new members-list temporary, and find out who is the owner,
       after that, we will send 'query' to the owner to get the newest members-list.
"""

from typing import Optional

from dimp import ID
from dimp import InstantMessage
from dimp import Content
from dimp import GroupCommand

from .history import GroupCommandProcessor


class ResetCommandProcessor(GroupCommandProcessor):

    def __temporary(self, sender: ID, members: list, group: ID) -> Optional[Content]:
        if self.contains_owner(members=members, group=group):
            facebook = self.facebook
            # temporary save
            if facebook.save_members(members=members, identifier=group):
                owner = facebook.owner(identifier=group)
                if owner is not None and owner != sender:
                    # NOTICE: to prevent counterfeit,
                    #         query the owner for newest member-list
                    cmd = GroupCommand.query(group=group)
                    self.messenger.send_content(content=cmd, receiver=owner)
            # response (no need to response this group command)
            return None
        else:
            # NOTICE: this is a partial member-list
            #         query the sender for full-list
            return GroupCommand.query(group=group)

    def __reset(self, new_members: list, group: ID) -> (list, list):
        facebook = self.facebook
        # existed members
        members: list = facebook.members(identifier=group)
        if members is None:
            members = []
        # removed member(s)
        remove_list = []
        for item in members:
            if item not in new_members:
                # found
                remove_list.append(item)
        # added member(s)
        add_list = []
        for item in new_members:
            if item not in members:
                # found
                add_list.append(item)
        # save changes
        if len(add_list) > 0 or len(remove_list) > 0:
            if not facebook.save_members(members=new_members, identifier=group):
                return None, None
        return add_list, remove_list

    #
    #   main
    #
    def process(self, content: Content, sender: ID, msg: InstantMessage) -> Optional[Content]:
        # assert isinstance(content, ResetCommand), 'group command error: %s' % content
        assert isinstance(content, GroupCommand), 'group command error: %s' % content
        facebook = self.facebook
        # new members
        new_members: list = self.members(content=content)
        if new_members is None or len(new_members) == 0:
            raise ValueError('reset group command error: %s' % content)
        group: ID = facebook.identifier(content.group)
        # 0. check whether group info empty
        if self.is_empty(group=group):
            # FIXME: group profile lost?
            # FIXME: how to avoid strangers impersonating group members?
            return self.__temporary(sender=sender, members=new_members, group=group)
        # 1. check permission
        if not facebook.is_owner(member=sender, group=group):
            if not facebook.exists_assistant(member=sender, group=group):
                raise AssertionError('only owner/assistant can reset: %s' % msg)
        # 2. reset
        added, removed = self.__reset(new_members=new_members, group=group)
        if added is not None:
            content['added'] = added
        if removed is not None:
            content['removed'] = removed
        # 3. response (no need to response this group command)
        return None


# register
GroupCommandProcessor.register(command=GroupCommand.RESET, processor_class=ResetCommandProcessor)



# -*- coding: utf-8 -*-
#
#   DIM-SDK : Decentralized Instant Messaging Software Development Kit
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Storage Protocol
    ~~~~~~~~~~~~~~~~

    Storage data (may be encrypted) by title for VIP users
"""

import json
from typing import Optional

from mkm.crypto.utils import base64_encode, base64_decode
from dimp import DecryptKey, SymmetricKey, ID
from dimp import Command


class StorageCommand(Command):
    """
        Storage Command
        ~~~~~~~~~~~~~~~

        data format: {
            type : 0x88,
            sn   : 123,

            command : "storage", // command name
            title   : "...",     // "contacts", "private_key", ...

            data    : "...",  // base64_encode(symmetric)
            key     : "...",  // base64_encode(asymmetric)
            //-- extra info
        }
    """

    STORAGE = 'storage'

    CONTACTS = 'contacts'
    PRIVATE_KEY = 'private_key'

    def __new__(cls, cmd: dict):
        """
        Create storage command

        :param cmd: command info
        :return: StorageCommand object
        """
        if cmd is None:
            return None
        elif cls is StorageCommand:
            if isinstance(cmd, StorageCommand):
                # return StorageCommand object directly
                return cmd
        # new StorageCommand(dict)
        return super().__new__(cls, cmd)

    def __init__(self, content: dict):
        if self is content:
            # no need to init again
            return
        super().__init__(content)
        # lazy
        self.__key: bytes = None
        self.__data: bytes = None
        self.__plaintext: object = None

    #
    #   Title
    #
    @property
    def title(self) -> str:
        string = self.get('title')
        if string is None or len(string) == 0:
            string = self.command
            assert string != self.STORAGE, 'storage command error: %s' % self
        return string

    #
    #   ID
    #
    @property
    def identifier(self) -> str:
        return self.get('ID')

    @identifier.setter
    def identifier(self, value: ID):
        if value is None:
            self.pop('ID', None)
        else:
            self['ID'] = value

    #
    #   Key (for decrypt data)
    #
    @property
    def key(self) -> Optional[bytes]:
        if self.__key is None:
            base64 = self.get('key')
            if base64 is not None:
                self.__key = base64_decode(base64)
        return self.__key

    @key.setter
    def key(self, value: bytes):
        if value is None:
            self.pop('key', None)
        else:
            self['key'] = base64_encode(value)
        self.__key = value

    #
    #   Data (encrypted)
    #
    @property
    def data(self) -> Optional[bytes]:
        if self.__data is None:
            base64 = self.get('data')
            if base64 is not None:
                self.__data = base64_decode(base64)
        return self.__data

    @data.setter
    def data(self, value: bytes):
        if value is None:
            self.pop('data', None)
        else:
            self['data'] = base64_encode(value)
        self.__data = value

    def decrypt(self, password: DecryptKey=None, private_key: DecryptKey=None) -> bytes:
        """
        Decrypt data
            1. decrypt key with private key (such as RSA) to a password
            2. decrypt data with password (symmetric key, such as AES, DES, ...)

        :param password:    symmetric key
        :param private_key: asymmetric private key
        :return: plaintext
        """
        if self.__plaintext is None:
            # get symmetric key
            key = None
            if password is not None:
                assert isinstance(password, SymmetricKey), 'password error: %s' % password
                key = password
            elif private_key is not None:
                # assert isinstance(private_key, PrivateKey), 'private key error: %s' % private_key
                key_data = private_key.decrypt(self.key)
                key = SymmetricKey(json.loads(key_data))
            # get encrypted data
            data = self.data
            if key is not None and data is not None:
                self.__plaintext = key.decrypt(data=data)
        return self.__plaintext

#
    #   Factories
    #
    @classmethod
    def new(cls, content: dict=None, title: str=None):
        """
        Create storage command

        :param content: command info
        :param title: title
        :return: StorageCommand object
        """
        if content is None:
            # create empty content
            content = {}
        # new StorageCommand(dict)
        if title is not None:
            if title == cls.CONTACTS:
                # compatible with v1.0
                return super().new(content=content, command=cls.CONTACTS)
            content['title'] = title
        return super().new(content=content, command=cls.STORAGE)


# register command class
Command.register(command=StorageCommand.STORAGE, command_class=StorageCommand)
Command.register(command=StorageCommand.CONTACTS, command_class=StorageCommand)
Command.register(command=StorageCommand.PRIVATE_KEY, command_class=StorageCommand)



# -*- coding: utf-8 -*-
#
#   DIM-SDK : Decentralized Instant Messaging Software Development Kit
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Address Name Service
    ~~~~~~~~~~~~~~~~~~~~

    A map for short name to ID, just like DNS
"""

from abc import ABC, abstractmethod
from typing import Optional

from mkm.address import ANYWHERE
from dimp import ID, ANYONE, EVERYONE

#
#   Founder
#
founder = ID.new(name='moky', address=ANYWHERE)
#
#   Reserved names
#
keywords = [
    "all", "everyone", "anyone", "owner", "founder",
    # --------------------------------
    "dkd", "mkm", "dimp", "dim", "dimt",
    "rsa", "ecc", "aes", "des", "btc", "eth",
    # --------------------------------
    "crypto", "key", "symmetric", "asymmetric",
    "public", "private", "secret", "password",
    "id", "address", "meta", "profile",
    "entity", "user", "group", "contact",
    # --------------------------------
    "member", "admin", "administrator", "assistant",
    "main", "polylogue", "chatroom",
    "social", "organization",
    "company", "school", "government", "department",
    "provider", "station", "thing", "robot",
    # --------------------------------
    "message", "instant", "secure", "reliable",
    "envelope", "sender", "receiver", "time",
    "content", "forward", "command", "history",
    "keys", "data", "signature",
    # --------------------------------
    "type", "serial", "sn",
    "text", "file", "image", "audio", "video", "page",
    "handshake", "receipt", "block", "mute",
    "register", "suicide", "found", "abdicate",
    "invite", "expel", "join", "quit", "reset", "query",
    "hire", "fire", "resign",
    # --------------------------------
    "server", "client", "terminal", "local", "remote",
    "barrack", "cache", "transceiver",
    "ans", "facebook", "store", "messenger",
    "root", "supervisor",
]


class AddressNameService(ABC):

    def __init__(self):
        super().__init__()
        # ANS records
        self.__caches = {
            'all': EVERYONE,
            'everyone': EVERYONE,
            'anyone': ANYONE,
            'owner': ANYONE,
            'founder': founder,
        }
        # reserved names
        reserved = {}
        for name in keywords:
            reserved[name] = True
        self.__reserved = reserved

    def is_reserved(self, name: str) -> bool:
        return self.__reserved.get(name)

    def cache(self, name: str, identifier: ID=None) -> bool:
        if self.is_reserved(name):
            # this name is reserved, cannot register
            return False
        if identifier is None:
            self.__caches.pop(name, None)
        else:
            self.__caches[name] = identifier
        return True

    def identifier(self, name: str) -> Optional[ID]:
        """ Get ID by short name """
        return self.__caches.get(name)

    def names(self, identifier: ID) -> Optional[list]:
        """ Get all short names with the same ID """
        array = []
        for (key, value) in self.__caches.items():
            if key == identifier:
                array.append(value)
        return array

    @abstractmethod
    def save(self, name: str, identifier: ID=None) -> bool:
        """
        Save ANS record

        :param name:       username
        :param identifier: user ID; if empty, means delete this name
        :return: True on success
        """
        if not self.cache(name=name, identifier=identifier):
            return False



# -*- coding: utf-8 -*-
#
#   DIM-SDK : Decentralized Instant Messaging Software Development Kit
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Apple Push Notification service (APNs)
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    A service for pushing notification to offline device
"""

from abc import ABC, abstractmethod

from apns2.client import APNsClient, NotificationPriority
from apns2.errors import APNsException
from apns2.payload import Payload


class ApplePushNotificationService:

    def __init__(self, credentials, use_sandbox=False, use_alternative_port=False, proto=None, json_encoder=None,
                 password=None, proxy_host=None, proxy_port=None):
        super().__init__()
        # APNs client parameters
        self.credentials = credentials
        self.use_sandbox = use_sandbox
        self.use_alternative_port = use_alternative_port
        self.proto = proto
        self.json_encoder = json_encoder
        self.password = password
        self.proxy_host = proxy_host
        self.proxy_port = proxy_port
        self.client = None  # APNsClient
        # topic
        self.topic = 'chat.dim.sechat'
        # delegate to get device token
        self.delegate = None  # IAPNsDelegate
        # counting offline messages
        self.badge_table = {}

    @staticmethod
    def info(msg: str):
        # Log.info('APNs:\t%s' % msg)
        pass

    @staticmethod
    def error(msg: str):
        # Log.error('APNs ERROR:\t%s' % msg)
        pass

    def badge(self, identifier: str) -> int:
        num = self.badge_table.get(identifier)
        if num is None:
            num = 1
        else:
            num = num + 1
        self.badge_table[identifier] = num
        return num

    def clear_badge(self, identifier: str) -> bool:
        if identifier in self.badge_table:
            self.badge_table.pop(identifier)
            return True

    def connect(self) -> bool:
        try:
            self.client = APNsClient(credentials=self.credentials, use_sandbox=self.use_sandbox,
                                     use_alternative_port=self.use_alternative_port,
                                     proto=self.proto, json_encoder=self.json_encoder, password=self.password,
                                     proxy_host=self.proxy_host, proxy_port=self.proxy_port)
            return True
        except IOError as error:
            self.error('failed to connect apple server: %s' % error)
            return False

    def send_notification(self, token_hex, notification, topic=None,
                          priority=NotificationPriority.Immediate, expiration=None, collapse_id=None) -> int:
        if self.client is None and self.connect() is False:
            self.error('cannot connect apple server, message dropped: %s' % notification)
            return -503  # Service Unavailable
        try:
            if topic is None:
                topic = self.topic
            # push to apple server
            self.client.send_notification(token_hex=token_hex, notification=notification, topic=topic,
                                          priority=priority, expiration=expiration, collapse_id=collapse_id)
            return 200  # OK
        except IOError as error:
            self.error('connection lost: %s' % error)
            return -408  # Request Timeout
        except APNsException as error:
            self.error('failed to push notification: %s, error %s' % (notification, error))
            return -400  # Bad Request

    def push(self, identifier: str, message: str) -> bool:
        # 1. check
        tokens = self.delegate.device_tokens(identifier=identifier)
        if tokens is None:
            self.error('cannot get device token for user %s' % identifier)
            return False
        # 2. send
        badge = self.badge(identifier)
        payload = Payload(alert=message, badge=badge, sound='default')
        success = 0
        for token in tokens:
            self.info('sending notification %s to user %s with token %s' % (message, identifier, token))
            # first try
            result = self.send_notification(token_hex=token, notification=payload)
            if result == -503:  # Service Unavailable
                # connection failed
                break
            elif result == -408:  # Request Timeout
                self.error('Broken pipe? try to reconnect again!')
                # reset APNs client
                self.client = None
                # try again
                result = self.send_notification(token_hex=token, notification=payload)
            if result == 200:  # OK
                success = success + 1
        if success > 0:
            self.info('sending notification success:%d badge=%d, %s' % (success, badge, identifier))
            return True


class IAPNsDelegate(ABC):
    """
        APNs Delegate
        ~~~~~~~~~~~~~
    """

    @abstractmethod
    def device_tokens(self, identifier: str) -> list:
        """ get device tokens in hex format """
        pass



# -*- coding: utf-8 -*-
#
#   DIM-SDK : Decentralized Instant Messaging Software Development Kit
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================


"""
    Notification Center
    ~~~~~~~~~~~~~~~~~~~

    Notification dispatcher
"""

from typing import Any
from weakref import WeakSet

from .observer import Notification, Observer


class NotificationCenter:

    def __init__(self):
        super().__init__()
        self.__observers: dict = {}

    def add(self, observer: Observer, name: str):
        """
        Add observer with notification name

        :param observer: notification observer
        :param name:     notification name
        :return:
        """
        array: WeakSet = self.__observers.get(name)
        if array is None:
            array = WeakSet()
            self.__observers[name] = array
        else:
            # check duplicated
            for item in array:
                if item == observer:
                    # already exists
                    return
        array.add(observer)

    def __remove(self, observer: Observer, name: str):
        array: WeakSet = self.__observers.get(name)
        if array is None:
            return
        # check
        for item in array:
            if item == observer:
                # got it
                array.remove(item)
                break

    def remove(self, observer: Observer, name: str=None):
        """
        Remove observer from notification center, no mather what names

        :param observer: notification observer to remove
        :param name:     notification name (if empty, remove from all names)
        :return:
        """
        if name is None:
            keys = self.__observers.keys()
            for item in keys:
                self.__remove(observer=observer, name=item)
        else:
            self.__remove(observer=observer, name=name)

    def post(self, notification: Notification=None,
             name: str=None, sender: Any=None, info: dict=None):
        """
        Post a notification (with name, sender and extra info)

        :param notification: notification object
        :param name:         notification name  (when 'notification' empty)
        :param sender:       notification sender(when 'notification' empty)
        :param info:         extra info         (when 'notification' empty)
        :return:
        """
        if notification is None:
            assert name is not None, 'Notification name empty'
            assert sender is not None, 'Notification sender empty'
            notification = Notification(name=name, sender=sender, info=info)
        # temporary array buffer, used as a snapshot of the state of current observers
        array = self.__observers.get(notification.name)
        if array is not None:
            # call observers one by one
            for observer in array:
                observer.received_notification(notification=notification)

    @classmethod
    def default(cls):
        return default_center


#
#   Singleton
#
default_center = NotificationCenter()



# -*- coding: utf-8 -*-
#
#   DIM-SDK : Decentralized Instant Messaging Software Development Kit
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Chatroom
    ~~~~~~~~

    Big group with admins
"""

from abc import abstractmethod
from typing import Optional

from dimp import ID, NetworkID
from dimp import Group, GroupDataSource


class ChatroomDataSource(GroupDataSource):
    """This interface is for getting information for chatroom

        Chatroom Data Source
        ~~~~~~~~~~~~~~~~~~~~

        Chatroom admins should be set complying with the consensus algorithm
    """

    @abstractmethod
    def admins(self, identifier: ID) -> Optional[list]:
        """
        Get all admins in the chatroom

        :param identifier: chatroom ID
        :return: admin ID list
        """
        pass


class Chatroom(Group):

    def __init__(self, identifier: ID):
        super().__init__(identifier=identifier)
        assert identifier.type == NetworkID.Chatroom, 'Chatroom ID type error: %s' % identifier

    @Group.delegate.getter
    def delegate(self) -> Optional[ChatroomDataSource]:
        facebook = super().delegate
        assert facebook is None or isinstance(facebook, ChatroomDataSource), 'error: %s' % facebook
        return facebook

    # @delegate.setter
    # def delegate(self, value: ChatroomDataSource):
    #     super(Chatroom, Chatroom).delegate.__set__(self, value)

    @property
    def admins(self) -> Optional[list]:
        return self.delegate.admins(identifier=self.identifier)
